#ifndef FRAMECPP__COMMON__FrameSpec_HH
#define FRAMECPP__COMMON__FrameSpec_HH

#if ! defined(SWIGIMPORTED)
#include <iostream>
#include <memory>
#include <sstream>
#include <stdexcept>
#endif /* ! defined(SWIGIMPORTED) */

#include "general/unordered_map.hh"
#include "general/SharedPtr.hh"
#include "general/types.hh"

namespace FrameCPP
{
  namespace Common
  {
    class IFrameStream;
    class IStream;
    class OStream;
    class Description;
    class StreamBase;
    class Verify;

    class FrameSpec
    {
    public:
      typedef INT_2U	version_type;
      typedef INT_2S	class_type;
      typedef INT_8U	size_type;
      typedef INT_4U	instance_type;

      static const CHAR_U MINOR_VERSION_BETA = 255;

      class Object;

      class ObjectInterface
      {
      public:
	typedef General::SharedPtr< Object >	object_type;
	typedef object_type demote_arg_type;
	typedef object_type demote_ret_type;
	typedef object_type promote_arg_type;
	typedef object_type promote_ret_type;


	virtual ~ObjectInterface( );

	/// \brief Return the number of bytes for this object
	virtual size_type Bytes( const StreamBase& Stream ) const = 0;

	/// \brief constructor - virtual
	virtual Object* Create( ) const;

	/// \brief constructor - virtual
	virtual Object* Create( IStream& Stream ) const = 0;

	/// \brief comparison operator
	virtual bool operator==( const Object& Obj ) const = 0;

	/// \brief constructor - virtual copy
	virtual Object* Clone( ) const;

	//---------------------------------------------------------------
	/// \brief Down grade an object
	///
	/// \param[in] Target
	///      This is the version of the frame specification of
	///      the demoted object.
	/// \param[in] Obj
	///      The base object.
	/// \param[in] Stream
	///      Stream from which Obj was created.
	///
	/// \return
	///      A pointer to the objected derived by demoting Obj.
	//---------------------------------------------------------------
	static object_type
	DemoteObject(  INT_2U Target, object_type Obj, Common::IStream* Stream );

	/// \brief Get FrSH/FrSE description of object
	virtual  const Description* GetDescription( ) const = 0;

	//---------------------------------------------------------------
	/// \brief Retrieve the class name of the object
	///
	/// \return
	///     The name of the class associated with the object.
	//---------------------------------------------------------------
	virtual const char* const ObjectStructName( ) const = 0;

	//---------------------------------------------------------------
	/// \brief Upgrade an object
	///
	/// \param[in] Target
	///     Promote Obj to this version of the frame specification.
	/// \param[in] Source
	///     Obj conforms to this version of the frame specification.
	/// \param[in] Obj
	///     Object to be converted
	/// \param[in] Stream
	///     Originating stream for Obj.
	///
	/// \return
	///     A pointer to the objected derived by promoting Obj.
	//---------------------------------------------------------------
	static object_type
	PromoteObject( INT_2U Target, INT_2U Source,
		       object_type Obj, Common::IStream* Stream );
	
	template< typename PreviousType, typename CurrentType >
	static object_type
	PromoteObject( INT_2U FrameSpecification,
		       INT_2U Source,
		       object_type Obj,
		       IStream* Stream );

	virtual void VerifyObject( Verify& Verifier,
				   IFrameStream& Stream ) const;

	//---------------------------------------------------------------
	/// \brief Write object to the stream
	//---------------------------------------------------------------
	virtual void Write( OStream& Stream ) const = 0;

      protected:
	//---------------------------------------------------------------
	/// \brief Demotes object to previous version of the frame spec
	//---------------------------------------------------------------
	virtual demote_ret_type
	demote(  INT_2U Target, demote_arg_type Obj, IStream* Stream ) const = 0;

	//---------------------------------------------------------------
	/// \brief Promotes object to another version of the frame spec
	//---------------------------------------------------------------
	virtual promote_ret_type
	promote(  INT_2U Target, promote_arg_type Obj, IStream* Stream ) const = 0;

	/// \brief Retrieve the class name associated with the object
	virtual class_type getClass( ) const = 0;

      };

      class Object
	: public ObjectInterface
      {
      public:
	/// \brief constructor
	Object( class_type Class, const Description* Description );

	/// \brief copy constructor
	Object( const Object& Source );

	/// \brief destructor
	virtual ~Object( );

	class_type GetClass( ) const;

	/// \brief Get FrSH/FrSE description of object
	const Description* GetDescription( ) const;

	const Object& operator=( const Object& Source );

	OStream& WriteNext( OStream& Stream ) const;

      protected:
	virtual class_type getClass( ) const;

      private:
	const class_type		m_class;
	const Description* const	m_desc;
      };

      template < typename ObjectType, typename ChecksumType >
      class ObjectWithChecksum
	: public ObjectType
      {
      public:
	typedef ObjectWithChecksum< ObjectType, ChecksumType >
	self_type;

	//---------------------------------------------------------------
	/// \brief Constructor - Default
	//---------------------------------------------------------------
	ObjectWithChecksum( );

	//---------------------------------------------------------------
	/// \brief Constructor
	//---------------------------------------------------------------
	ObjectWithChecksum( const Description* Description );

	//---------------------------------------------------------------
	/// \brief Constructor
	//---------------------------------------------------------------
	ObjectWithChecksum( class_type Class, const Description* Description );

	//---------------------------------------------------------------
	/// \brief copy constructor
	//---------------------------------------------------------------
	ObjectWithChecksum( const ObjectWithChecksum& Source );

	//---------------------------------------------------------------
	/// \brief Constructor - Stream
	//---------------------------------------------------------------
	ObjectWithChecksum( IStream& Source );

	//---------------------------------------------------------------
	/// \brief destructor
	//---------------------------------------------------------------
	virtual ~ObjectWithChecksum( );

	//---------------------------------------------------------------
	/// \brief Return the number of bytes for this object
	//---------------------------------------------------------------
	virtual size_type Bytes( const StreamBase& Stream ) const;

	//---------------------------------------------------------------
	/// \brief Data class to be used in class field of FrSE
	//---------------------------------------------------------------
	static const char* CheckSumDataClass( );

	//---------------------------------------------------------------
	/// \brief Comment to be used in comment field of FrSE
	//---------------------------------------------------------------
	static const char* CheckSumDataComment( );

	virtual self_type* Create( Common::IStream& Stream ) const;
	//---------------------------------------------------------------
	/// \brief Write object to the stream
	//---------------------------------------------------------------
	virtual void Write( OStream& Stream ) const;

      protected:
	virtual size_type pBytes( const StreamBase& Stream ) const = 0;

	virtual self_type* pCreate( IStream& Stream ) const = 0;

	virtual void pWrite( OStream& Stream ) const = 0;

      private:
	/* using ObjectType::Create; */

	virtual Object* Create( ) const
	{
	  return ObjectType::Create( );
	}

	ChecksumType	m_chkSum;
      };

      class Info
      {
      public:
	typedef Verify verify_type;
	typedef General::SharedPtr< Object > object_type;

	enum frame_object_types {
	  FSI_PTR_STRUCT = -3,
	  FSI_COMMON_ELEMENTS = -2,
	  FSI_FR_HEADER = -1,
	  FSI_FR_NULL = 0,
	  FSI_FR_SH = 1,
	  FSI_FR_SE = 2,
	  FSI_FRAME_H,
	  FSI_FR_ADC_DATA,
	  FSI_FR_DETECTOR,
	  FSI_FR_END_OF_FILE,
	  FSI_FR_END_OF_FRAME,
	  FSI_FR_EVENT,
	  FSI_FR_HISTORY,
	  FSI_FR_MSG,
	  FSI_FR_PROC_DATA,
	  FSI_FR_RAW_DATA,
	  FSI_FR_SER_DATA,
	  FSI_FR_SIM_DATA,
	  FSI_FR_SIM_EVENT,
	  FSI_FR_STAT_DATA,
	  FSI_FR_SUMMARY,
	  FSI_FR_TABLE,
	  FSI_FR_TOC,
	  FSI_FR_VECT,
	  //-------------------------------------------------------------
	  // Backwards compatability
	  //-------------------------------------------------------------
	  FSI_FR_TRIG_DATA = FSI_FR_EVENT
	};

	typedef void (*verification_func_type)( const FrameSpec::Object *,
						verify_type&,
						IFrameStream& Stream );
	Info( version_type MajorVersion,
	      version_type MinorVersion,
	      verification_func_type VerificationFunc
	      = (verification_func_type)NULL );

	~Info( );

	inline verification_func_type
	VerificationFunc( ) const
	{
	  return m_verification_func;
	}

	inline version_type
	Version( ) const
	{
	  return m_version;
	}

	inline version_type
	VersionMinor( ) const
	{
	  return m_version_minor;
	}

	const Object* FrameObject( frame_object_types ObjectId ) const;

	void FrameObject( object_type ObjectTemplate );

	void FrameObject( object_type::element_type* ObjectTemplate );

      private:
	typedef General::unordered_map< version_type, object_type > object_container;

	object_container	m_definitions;
	version_type		m_version;
	version_type		m_version_minor;
	verification_func_type	m_verification_func;
	
      };

      static Info* SpecInfo( version_type Version );

      static void SpecInfo( version_type Version, Info* Info );

    private:
      class frame_spec_container
	: public General::unordered_map< version_type, Info* >
      {
      public:
	frame_spec_container( );
	virtual ~frame_spec_container( );
      };


      static frame_spec_container& frame_specs( );

      static void cleanup_frame_specs( );
    };

    //===================================================================
    //===================================================================
    inline void FrameSpec::Info::
    FrameObject( object_type::element_type* ObjectTemplate )
    {
      object_type	obj( ObjectTemplate );

      FrameObject( obj );
    }
    
    //===================================================================
    //===================================================================
    inline FrameSpec::class_type FrameSpec::Object::
    GetClass( ) const
    {
      return m_class;
    }

    /// \brief Get FrSH/FrSE description of object
    inline const Description* FrameSpec::Object::
    GetDescription( ) const
    {
      return m_desc;
    }

    inline const FrameSpec::Object& FrameSpec::Object::
    operator=( const Object& Source )
    {
      if ( Source.m_class != m_class )
      {
	throw
	  std::domain_error( "incompatable assignment of FrameSpec::Object" );
      }
      return *this;
    }
    //-------------------------------------------------------------------
    /// \brief 
    ///
    /// \param[in] FrameSpecification
    ///     The version of the current frame specification.
    /// \param[in] Source
    ///     The version of the frame specification to which Obj
    ///     was built
    /// \param[in] Obj
    ///	    The object to be converted.
    /// \param[in] Stream
    ///     The stream from which Obj was created
    ///
    /// \return
    ///     Either the Obj or NULL pointer wrapped in auto_ptr.
    //-------------------------------------------------------------------
    template< typename PreviousType, typename CurrentType >
    FrameSpec::ObjectInterface::promote_ret_type FrameSpec::ObjectInterface::
    PromoteObject( INT_2U FrameSpecification,
		   INT_2U Source,
		   FrameSpec::ObjectInterface::promote_arg_type Obj,
		   IStream* Stream )
    {
      if ( Source >= FrameSpecification )
      {
	//---------------------------------------------------------------
	// Nothing to do here
	//---------------------------------------------------------------
	return Obj;
      }
      //-----------------------------------------------------------------
      // Promote from the previous level
      //-----------------------------------------------------------------
      object_type retval;

      object_type
	previous_object( PreviousType::Promote( Source, Obj, Stream ) );
      if ( previous_object )
      {
	//--------------------------------------------------------------
	// Object is not NULL, need to upconvert
	//--------------------------------------------------------------
	General::SharedPtr< PreviousType >
	  previous
	  = General::StaticPointerCast< PreviousType >( previous_object );

	retval.reset( new CurrentType( *previous, Stream ) );
      }

      return retval;
    }

    //===================================================================
    // Tempalted functions
    //===================================================================
    template < typename T >
    FrameSpec::size_type Bytes( const T& Data );
  
    template< typename T >
    void VerifyObject( const T& Obj,
		       FrameSpec::Info::verify_type& Verifier,
		       IFrameStream& Stream );

    //-------------------------------------------------------------------
    /// \brief 
    ///
    /// This routine handles the case were an object does not exist in
    /// previous version of the frame specification.
    ///
    /// \param[in] FrameSpecification
    ///     The version of the current frame specification.
    /// \param[in] Source
    ///     The version of the frame specification to which Obj
    ///     was built
    /// \param[in] Obj
    ///	  The object to be converted.
    ///
    /// \return
    ///     Either the Obj or NULL pointer wrapped in auto_ptr.
    //-------------------------------------------------------------------
    inline FrameSpec::ObjectInterface::promote_ret_type
    PromoteObject( INT_2U FrameSpecification,
		   INT_2U Source,
		   FrameSpec::ObjectInterface::promote_arg_type Obj )
    {
      FrameSpec::ObjectInterface::promote_ret_type	retval;

      if ( Source >= FrameSpecification )
      {
	retval = Obj;
      }
      return retval;
    }

  } // namespace - Common

  typedef Common::FrameSpec::size_type cmn_streamsize_type;

  typedef Common::FrameSpec::Object Object;

  class Unimplemented
    : public std::exception
  {
  public:
    inline
    Unimplemented( const std::string& Call, CHAR_U FrameSpecVersion,
		   const char* Filename, int Line )
    {
      std::ostringstream       m;

      m << "The call \"" << Call << "\""
	<< " at line " << Line
	<< " of file " << Filename
	<< " is currently not implemented for version "
	<< INT_2U( FrameSpecVersion )
	<< " of the frame specification."
	;
      msg = m.str( );
    }

    inline virtual
    ~Unimplemented( )
      throw( )
    {
    }

    inline virtual const char*
    what( ) const
      throw( )
    {
      return msg.c_str( );
    }

  private:
    std::string	msg;
  };

  template< typename T >
  inline bool
  compare( const T& LHS, const Common::FrameSpec::Object& Obj )
  {
    bool retval = ( &LHS == &Obj );

    if ( retval == false )
    {
      const T* rhs;
      if ( ( LHS.GetClass( ) == Obj.GetClass( ) )
	   && ( ( rhs = dynamic_cast< const T* >( &Obj ) ) ) )
      {
	retval = ( LHS == *rhs );
      }
    }
    return retval;
  }

  template< typename T >
  inline bool
  IsOfType( Common::FrameSpec::Object* Obj )
  {
    if ( dynamic_cast< T* >( Obj ) == (T*)NULL )
    {
      return false;
    }
    return true;
  }

} // namespace - FrameCPP

#endif /* FRAMECPP__COMMON__FrameSpec_HH */
